/************************************************************************
*Copyright (C) 2003 Elliott Michell					*
*Elliott Mitchell takes no responsibility for failures, or damage done	*
*by this program.							*
*									*
*$Id: lzss.c,v 1.8 2003/11/13 01:50:48 ehem Rel $			*
*************************************************************************
*An implementation of a varient of the well known LZSS compression	*
*algorithm. LZSS is a varient of LZ77 (Lempel-Ziv, 1977), sometimes	*
*refered to as "sliding window" compression.				*
*									*
*LZ77 encoding is very simple. The encoding consists of tokens		*
*containing three values; an index into the window, the length of the	*
*match and the next token/character of the data. Decoding is very	*
*simple, starting at the specified <index> into the output buffer copy	*
*<length> tokens to the current position and then output the next	*
*token. Encoding is also pretty simple, look for a position in the	*
*output buffer that matches any data in the input buffer, and then	*
*output the next piece of data. Note that the input and output buffers	*
*need to be joined, as a match can start in the output buffer and reach	*
*all the way forward into the input buffer (the two together form a	*
*window into the data, hence sliding window compression).		*
*									*
*Well, one really big gotcha, when starting out the output buffer is	*
*empty and hence we'll have repeated tokens with matches of length zero	*
*until at least one character recurs. A smaller gotcha in the formal	*
*definition of LZ77 is that linear searching was used when searching	*
*for a match in the window. Well, move forward some time, and someone	*
*comes up with the trivial solutions. First, use a single bit to	*
*indicate whether the following is a raw data byte, or a full token	*
*with a match. In doing so the worst case goes from tripling the size	*
*of the first 256 bytes (matches of zero, every token encoding 1 byte	*
*of data) and fifty percent increase afterwords (matches of one, every	*
*token encoding 2 bytes of data), to merely adding one byte for every	*
*eight bytes of data. Second, use a tree for searching the window,	*
*rather than using a linear search.					*
*									*
*Well, but these two trivial innovations together and you have LZSS.	*
*The later (and other than having the same creators, completely		*
*unrelated) compression algorithm LZ78 shared similar weaknesses. Well	*
*the solutions are nearly the same, a hash table rather than tree,	*
*otherwise they're in fact identical. Well, trivial refinements still	*
*get a new name and that was LZW. The big difference is LZW got		*
*patented, despite both of these being extremely trivial and obvious	*
*refinements.								*
*									*
*Okay, enough of the formal definitions, now to what is done here.	*
*									*
*We use something akin to a single-bit for encoding match versus raw	*
*data. Instead of an actual bit though, we simply simply allocate a	*
*range of values for data and a range for matches (BNDTOK, can be	*
*overriden). Since unmatched bytes are likely to occur in multiples and	*
*we have a range of values available we use that to designate multiple	*
*unmatched characters at once. Further, since designating zero		*
*unmatched bytes is useless, we designate that we will encode a minimum	*
*of one byte (MINSTR, can be overriden). Also note that given the	*
*pattern of match-unmatch-match, the code cannot detect the possibility	*
*of the first two being small and therefore worth combining because the	*
*unmatch otherwise stealing from the second match.			*
*									*
*Seeing how having a larger number of places for matches is likely more	*
*important than longer matches (depending upon the source this could be	*
*wrong, but I suspect this is likelier for the general case), we use	*
*the range data as the length and then a following byte as the index	*
*into the window. As we encode unmatched strings in the above fashion,	*
*encoding matches of length zero is obviously useless. The minimum	*
*length match to consider (MINMAT, can be overriden) is a bit trickier	*
*than this. If we have a match of length two, this can be encoded using	*
*one of these tokens using two bytes, this is shorter than the three	*
*bytes used for a non-match (one byte length/flag, and two of data).	*
*This assumes we are following a match token though. If we are		*
*following an unmatched token, those two bytes can simply be merged	*
*with the prior token and we've shrunk the maximum possible match. This	*
*also improves the performance of the decoder as there will tend to be	*
*fewer tokens per unit of data, however this means longer unmatched	*
*strings will be broken up by running into the maximum length rather	*
*than being broken by tiny matches for free.				*
*									*
*Well, the final ingredient is the tree. Just for the heck of it we use	*
*Glib's GTree. Unfortunatly given the results I suspect Glib is using	*
*AVL trees (*bad* performance, might try messing with Glib's memory	*
*routinues to see if that is the problem). AVL trees are pretty		*
*efficient for lookups, but insert/delete are very expensive; Red-Black	*
*trees would be better here. (as you'll notice, we do a /maximum/ of	*
*one lookup for every insert/delete, at a minimum it is one lookup for	*
*over a hundred inserts/deletes)					*
*									*
*Lastly, compression performance. Useful. This can't hope to seriously	*
*compete with a full library like libz or libbz2, but for the general	*
*case it does do a useful level of compression.				*
************************************************************************/


#include <stdio.h>
#include <glib.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>


/************************************************************************
*The (compile time) changable values:					*
*									*
*BNDTOK: The boundary between unmatch and match ranges. This designates	*
*the lowest match value. Higher and matches will be shorter, but if	*
*there are long unmatched sections it might work better.		*
*									*
*MINSTR: The minimum unmatched string we consider. With noiser data	*
*having longer non-matches may be more valuble than allowing efficient	*
*encoding of short unmatched strings. This must be at least one, and I	*
*doubt anything greater than three is worthwhile.			*
*									*
*MINMAT: The minimum possible match. Larger makes longer matches	*
*possible, but makes smaller matches impossible. Two is a useful value,	*
*but any smaller value is worthless (must be at least one). Values	*
*larger than three will tend to lose by throwing away more common small	*
*matches.								*
************************************************************************/

#ifndef BNDTOK
#define BNDTOK 128
#endif

#ifndef MINSTR
#define MINSTR 1
#endif
#ifndef MINMAT
#define MINMAT 3
#endif

#define MAXSTR (MINSTR+(BNDTOK-1))
#define MAXMAT (MINMAT+(0xFF-BNDTOK))

#define MAXVAL 0xFF

#define BUFHALF PAGESIZE
#define BUFFULL (PAGESIZE+PAGESIZE)


/*Hrmm, I thought this was a macro*/
size_t PAGESIZE;

char *buf;
int bufi, bufcnt;

struct matchstats {
	const char *index;
	int size;
};

gint entrcmp(const char *, const char *, struct matchstats *);


int main(int argc, char **argv, char **envp)
{
	FILE *in, *out;
	char *func;
	func=*argv+strlen(*argv);
	while(func[-1]!='/'&&func>*argv) --func;


	PAGESIZE=sysconf(_SC_PAGESIZE);
	if(!(buf=(char *)memalign(BUFHALF, BUFFULL))) exit(-1);

	bufi=0;
	in=stdin;
	out=stdout;

	if(strcmp("unlzss", func)) {
		GTree *table;
		struct matchstats stat;
		int strcnt;
		int tmp;

		if(!(table=g_tree_new_with_data((GCompareDataFunc)entrcmp,
(gpointer)&stat))) exit(-1);
		strcnt=0;

		setvbuf(in, NULL, _IONBF, 0);
		bufcnt=fread(buf, 1, BUFFULL, in);

		do {
			if(bufcnt>=BUFFULL&&bufi>bufcnt-MAXSTR-MAXMAT) {
				bufi-=BUFHALF;
				if(fseek(in, -BUFHALF, SEEK_SET)) {
					bufcnt-=BUFHALF;
					g_memmove(buf+bufi-MAXSTR-MAXVAL,
buf+bufi+BUFHALF-MAXSTR-MAXVAL, BUFHALF-bufi+MAXSTR+MAXVAL);
					bufcnt+=fread(buf+BUFHALF, 1,
BUFFULL-bufcnt, in);
				} else {
					/*File, possible mmap() I/O*/
					bufcnt=fread(buf, 1, BUFFULL, in);
				}
			}

			stat.size=0;
			g_tree_lookup(table, buf+bufi); /*Yeah, we discard it*/
			if(stat.size>=MINMAT) { /*A worthy match!*/
				if(strcnt) {
					if(putc(strcnt-MINSTR, out)==EOF)
exit(1);
					if(fwrite(buf+bufi-strcnt, 1, strcnt,
out)!=strcnt) exit(1);
					strcnt=0;
				}
				/*Buffer end could be interesting on EOF*/
				if(bufi+stat.size>bufcnt) tmp=bufcnt-bufi;
				else tmp=stat.size;
				tmp+=BNDTOK-MINMAT;
				if(putc(tmp, out)==EOF) exit(1);
				tmp=buf+bufi-stat.index-1;
				if(tmp<0) {
					tmp+=BUFHALF; /*wraparound*/
				}
				if(putc(tmp, out)==EOF) exit(1);

				tmp=stat.size;
				do {
					g_tree_steal(table, buf+bufi-MAXVAL);
					g_tree_replace(table,buf+bufi,buf+bufi);
					++bufi;
				} while(--tmp);
			} else {
#if MINSTR != 1
				if(strcnt) {
#endif
				g_tree_steal(table, buf+bufi-MAXVAL);
				g_tree_replace(table, buf+bufi, buf+bufi);
				++bufi;
				if(++strcnt>=MAXSTR) {
					if(putc(strcnt-MINSTR, out)==EOF)
exit(1);
					if(fwrite(buf+bufi-strcnt, 1, strcnt,
out)!=strcnt) exit(1);
					strcnt=0;
				}
#if MINSTR != 1
				} else {
					strcnt=tmp=MINSTR;
					do {
						g_tree_steal(table,
buf+bufi-MAXVAL);
						g_tree_replace(table, buf+bufi,
buf+bufi);
						++bufi;
					} while(--tmp);
				}
#endif
			}
		} while(bufi<bufcnt);

		if(strcnt) {
			if(putc(strcnt-MINSTR, out)==EOF) exit(1);
			if(fwrite(buf+bufi-strcnt, 1, strcnt, out)!=strcnt)
exit(1);
		}
		g_tree_destroy(table);
	} else {
		int cnt, tok;

		setvbuf(out, NULL, _IONBF, 0);

		while((cnt=getc(in))!=EOF) {
			if(bufi>=BUFHALF+MAXVAL) {
				if(fwrite(buf, 1, BUFHALF, out)!=BUFHALF)
exit(1);
				bufi-=BUFHALF;
				g_memmove(buf, buf+BUFHALF, bufi);
			}
			if(cnt<BNDTOK) {
				cnt+=MINSTR;
				bufi+=fread(buf+bufi, 1, cnt, in);
			} else {
				if((tok=getc(in))==EOF) break;
				cnt+=MINMAT-BNDTOK;
/*				g_memmove(buf+bufi, (buf+bufi-1)-tok, cnt);
				bufi+=cnt;*/
				do {
					buf[bufi]=buf[bufi-1-tok];
					++bufi;
				} while(--cnt);
			}
		}
		if(fwrite(buf, 1, bufi, out)!=bufi) exit(1);
	}
	return(0);
}

gint entrcmp(const char *a, const char *b, struct matchstats *stat)
{
	int index=0;

	if(a>buf+bufi) a-=BUFHALF;
	if(b>buf+bufi) b-=BUFHALF;

	while(a[index]==b[index]) if(++index>=MAXMAT) break;

	if(index>stat->size) {
		if(a<b) stat->index=a;
		else stat->index=b;
		stat->size=index;
	}

	return(a[index]-b[index]);
}

